@@ -14,6 +14,7 @@ from spnego._context import (
     GSSMech,
     IOVUnwrapResult,
     IOVWrapResult,
+    SecPkgContextSizes,
     UnwrapResult,
     WinRMWrapResult,
     WrapResult,
@@ -59,11 +60,11 @@ GSSAPI_IOV_IMP_ERR = None
 try:
     from gssapi.raw import IOV as GSSIOV
     from gssapi.raw import IOVBuffer as GSSIOVBuffer
-    from gssapi.raw import IOVBufferType, unwrap_iov, wrap_iov
-except ImportError as err:
-    GSSAPI_IOV_IMP_ERR = sys.exc_info()
+    from gssapi.raw import IOVBufferType, unwrap_iov, wrap_iov, wrap_iov_length
+except ImportError as e:
+    GSSAPI_IOV_IMP_ERR = str(e)
     HAS_IOV = False
-    log.debug("Python gssapi IOV extension not available: %s" % str(GSSAPI_IOV_IMP_ERR[1]))
+    log.debug("Python gssapi IOV extension not available: %s" % GSSAPI_IOV_IMP_ERR)
 
 _GSS_C_INQ_SSPI_SESSION_KEY = "1.2.840.113554.1.2.2.5.5"
 
@@ -144,16 +145,16 @@ def _get_gssapi_credential(
 
             if isinstance(cred, KerberosKeytab):
                 username = cred.principal or ""
-                password = cred.keytab
+                password = to_bytes(cred.keytab)
                 is_keytab = True
             else:
                 username = cred.username
-                password = cred.password
+                password = _encode_kerb_password(cred.password)
                 is_keytab = False
 
             raw_cred = _kinit(
                 to_bytes(username),
-                to_bytes(password),
+                password,
                 forwardable=forwardable,
                 is_keytab=is_keytab,
             )
@@ -292,6 +293,40 @@ def _gss_acquire_cred_from_ccache(
         return gssapi_creds
 
 
+def _encode_kerb_password(
+    value: str,
+) -> bytes:
+    """Encode string to use for Kerberos passwords.
+
+    Encodes the input string to use with Kerberos functions as a password. This
+    is a special encoding method to ensure that any invalid surrogate chars are
+    encoded as the replacement char U+FFFD. This is needed when dealing with
+    randomly generated passwords like gMSA or machine accounts. The raw UTF-16
+    bytes can be encoded in a string with the following:
+
+        b"...".decode("utf-16-le", errors="surrogatepass")
+
+    The invalid surrogate pairs in the UTF-16 byte sequence will be preserved
+    in the str value allowing this function to replace it as needed. This
+    means the value can be used with both NTLM and Kerberos authentication with
+    the same value.
+
+    Args:
+        value: The string to encode to bytes.
+
+    Returns:
+        bytes: The encoded string value.
+    """
+    b_data = []
+    for c in value:
+        try:
+            b_data.append(c.encode("utf-8", errors="strict"))
+        except UnicodeEncodeError:
+            b_data.append(b"\xef\xbf\xbd")
+
+    return b"".join(b_data)
+
+
 class GSSAPIProxy(ContextProxy):
     """GSSAPI proxy class for GSSAPI on Linux.
 
@@ -312,7 +347,6 @@ class GSSAPIProxy(ContextProxy):
         options: NegotiateOptions = NegotiateOptions.none,
         **kwargs: typing.Any,
     ) -> None:
-
         if not HAS_GSSAPI:
             raise ImportError("GSSAPIProxy requires the Python gssapi library: %s" % GSSAPI_IMP_ERR)
 
@@ -444,6 +478,19 @@ class GSSAPIProxy(ContextProxy):
 
         return out_token
 
+    @wrap_system_error(NativeError, "Getting context sizes")
+    def query_message_sizes(self) -> SecPkgContextSizes:
+        if not self._context:
+            raise NoContextError(context_msg="Cannot get message sizes until context has been established")
+
+        iov = GSSIOV(
+            IOVBufferType.header,
+            b"",
+            std_layout=False,
+        )
+        wrap_iov_length(self._context, iov)
+        return SecPkgContextSizes(header=len(iov[0].value or b""))
+
     @wrap_system_error(NativeError, "Wrapping data")
     def wrap(self, data: bytes, encrypt: bool = True, qop: typing.Optional[int] = None) -> WrapResult:
         if not self._context:
@@ -544,6 +591,7 @@ class GSSAPIProxy(ContextProxy):
             (ContextReq.sequence_detect, "out_of_sequence_detection"),
             (ContextReq.confidentiality, "confidentiality"),
             (ContextReq.integrity, "integrity"),
+            (ContextReq.dce_style, "dce_style"),
             # Only present when the DCE extensions are installed.
             (ContextReq.identify, "identify"),
             # Only present with newer versions of python-gssapi https://github.com/pythongssapi/python-gssapi/pull/218.
@@ -571,4 +619,11 @@ class GSSAPIProxy(ContextProxy):
             auto_alloc = [BufferType.header, BufferType.padding, BufferType.trailer]
             buffer_alloc = buffer.type in auto_alloc
 
-        return GSSIOVBuffer(IOVBufferType(buffer.type), buffer_alloc, buffer_data)
+        buffer_type = buffer.type
+        if buffer.type == BufferType.data_readonly:
+            # GSSAPI doesn't have the SSPI equivalent of SECBUFFER_READONLY.
+            # the GSS_IOV_BUFFER_TYPE_EMPTY seems to produce the same behaviour
+            # so that's going to be used instead.
+            buffer_type = BufferType.empty
+
+        return GSSIOVBuffer(IOVBufferType(buffer_type), buffer_alloc, buffer_data)
